#!/usr/bin/env python
###########################################################################
# cgkit - Python Computer Graphics Kit
# Copyright (C) 2004 Matthias Baas (baas@ira.uka.de)
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# cgkit homepage: http://cgkit.sourceforge.net
###########################################################################

from distutils.core import setup, Extension
from distutils.sysconfig import get_python_lib
import sys, shutil, os, os.path, time, glob, stat

########################## Functions #################################

# isNewer
def isNewer(file1, file2):
    """Check if file1 is newer than file2.

    Returns True if file1 is newer than file2 or if file2 doesn't exist.
    file1 must be an existing file.
    """
    if not os.path.exists(file2):
        return True
    return os.stat(file1)[stat.ST_MTIME]>os.stat(file2)[stat.ST_MTIME]

# packageDLL
def packageDLL(dll):
    """Include a DLL into the package.

    The DLL is copied and will be made available in the package directory.

    dll: Full name of the DLL.
    """
    global data_files
    dllname = os.path.basename(dll)
    # Copy the DLL (only if the source is newer than the dest)...
    dstdll = os.path.join(PACKAGE_NAME, dllname)
    if (not os.path.exists(dstdll)) or (os.path.getmtime(dstdll)<os.path.getmtime(dll)):
        print 'Copying DLL %s...'%dll
        shutil.copyfile(dll, dstdll)
    # Add it to the list of data files...
    dest = os.path.join(get_python_lib(plat_specific=True), PACKAGE_NAME)
    n = len(os.path.commonprefix([sys.exec_prefix, dest]))
    dest = dest[n+1:]
    data_files += [(dest, [dll])]

# enabledStr
def enabledStr(flag):
    if flag:
        return "enabled"
    else:
        return "disabled"

# updateInfoModule
def updateInfoModule(cgkit_light):
    """Updates the module cgkitinfo.

    Updates the info module so that the version string contains the
    current date and time.
    """

    print "Updating version in the cgkitinfo module..."
    template = os.path.join("cgkit", "cgkitinfo_template.txt")
    infomod = os.path.join("cgkit", "cgkitinfo.py")

    # Read the cgkitinfo template module
    try:
        f = file(template)
        lines = f.readlines()
        f.close()
    except:
        print 'Error: Could not read file "%s"'%template
        sys.exit(1)

    # Replace the version string...
    version_info = None
    for i in range(len(lines)):
        s = lines[i]
        f = s.split(" ")
        if len(f)==0:
            continue
        if f[0]=="version_info":
            try:
                exec s.strip()
                if type(version_info)!=tuple:
                    raise Exception
                if len(version_info)!=4:
                    raise Exception
                major,minor,micro,rlevel = version_info
                if type(major)!=int or type(minor)!=int or type(micro)!=int:
                    raise Exception
                if type(rlevel)!=str:
                    raise Exception
            except:
                print 'Invalid version tuple in file "%s": %s'%(template,s.strip())
                sys.exit(1)
        if f[0]=="version":
            if version_info==None:
                print '%s: version_info must occur before version'%template
                sys.exit(1)
            major,minor,micro,rlevel = version_info
            v = "%d.%d.%d"%(major,minor,micro)
            if rlevel!="final":
                v+=rlevel
            if cgkit_light:
                v += " 'light'"
            v+=" (%s)"%time.strftime("%b %d %Y, %H:%M")
            lines[i] = 'version = "%s"\n'%v
            print "Version:",v

    lines.append("\ncgkit_light = %s"%cgkit_light)

    # Save the new content to the actual module file...
    try:
        f = file(infomod, "wt")
        f.writelines(lines)
        f.close()
    except:
        print 'Could not write file "%s"'%infomod
        sys.exit(1)

######################################################################
######################################################################
######################################################################

# Name of the entire package (you can't use just this variable to rename
# the package, all the directories and source files have to be updated
# accordingly)
PACKAGE_NAME = "cgkit"

# The following variables are used to customize the compile process

INC_DIRS = []
LIB_DIRS = []
LIBS     = []
CC_ARGS  = []
LINK_ARGS = []
MACROS    = []
data_files = []
scripts = []

# Default values
INSTALL_CGKIT_LIGHT = False
LIB3DS_AVAILABLE = False
CYBERX3D_AVAILABLE = False
OGRE_AVAILABLE = False
THREEDXWARE_AVAILABLE = False
WINTAB_AVAILABLE = False
GLOVESDK_AVAILABLE = False
GLOVESDK_BASE = None
BOOST_BASE = None
BOOST_DLL = "boost_python.dll"
USING_STLPORT = False
STLPORT_BASE = ""
STLPORT_DLL = "stlport_vc646.dll"
LIB3DS_PATH = None
LIB3DS_LIBNAME = "3ds"
CYBERX3D_PATH = None
CYBERX3D_LIB = None   # Default: "CyberX3D" (Windows) / "cx3d" (other)
CYBERX3D_XERCES_LIB = None  # Default: "xerces-c_2" (Windows) / "xerces-c" (other)
CYBERX3D_XERCES_DLL = "xerces-c_2_5_0.dll"
XERCES_PATH = None

# Read the config file
if os.path.exists("config.cfg"):
    execfile("config.cfg")
else:
    print 70*"-"
    print "Warning: No config file available (config.cfg)"
    print 70*"-"

# Disable all optional libs, STL and Boost for the light version...
if INSTALL_CGKIT_LIGHT:
    LIB3DS_AVAILABLE = False
    CYBERX3D_AVAILABLE = False
    OGRE_AVAILABLE = False
    THREEDXWARE_AVAILABLE = False
    WINTAB_AVAILABLE = False
    GLOVESDK_AVAILABLE = False
    USING_STLPORT = False
    BOOST_BASE = None

# Normalize the paths specified by the config file
if BOOST_BASE!=None:
    BOOST_BASE = os.path.normpath(BOOST_BASE)
if LIB3DS_PATH!=None:
    LIB3DS_PATH = os.path.normpath(LIB3DS_PATH)
if CYBERX3D_PATH!=None:
    CYBERX3D_PATH = os.path.normpath(CYBERX3D_PATH)
if XERCES_PATH!=None:
    XERCES_PATH = os.path.normpath(XERCES_PATH)

# Settings common for all platforms

if BOOST_BASE==None:
    BOOST_BIN = None
else:
    BOOST_BIN = os.path.join(BOOST_BASE, "libs", "python", "build", "bin-stage")

# Locations of include files
if BOOST_BASE!=None:
    INC_DIRS = [BOOST_BASE]+INC_DIRS
INC_DIRS = [os.path.join("supportlib","include")]+INC_DIRS
 
# Locations of library files
if BOOST_BIN!=None:
    LIB_DIRS = [BOOST_BIN] + LIB_DIRS
LIB_DIRS = [os.path.join("supportlib","lib")] + LIB_DIRS

# Libraries to link with
if sys.platform!="darwin":
	LIBS += ["boost_python"]
LIBS += ["core"]

# Scripts
scripts = ["viewer.py", "render.py", "utilities/postbake.py"]

# Lib3ds
if LIB3DS_AVAILABLE:
#    INC_DIRS += [LIB3DS_PATH]
#    LIB_DIRS += [os.path.join(LIB3DS_PATH, "msvc", "Build", "Release")]
    LIBS += [LIB3DS_LIBNAME]
    MACROS += [("LIB3DS_AVAILABLE", None)]

# CyberX3D
if CYBERX3D_AVAILABLE:
    MACROS += [("CYBERX3D_AVAILABLE", None)]

# Ogre3D
if OGRE_AVAILABLE:
    LIBS += ["OgreMain"]
    MACROS.append(("OGRE_AVAILABLE", None))
    scripts += ["viewerOgre.py"]

######################################################################
# Do some checks...

# Check if the boost/python.hpp header can be found...
if BOOST_BASE!=None and (not os.path.exists(os.path.join(BOOST_BASE, "boost", "python.hpp"))):
    print 70*"-"
    print "Error: The header <boost/python.hpp> is not available."
    print "Please check the BOOST_BASE variable in your config file."
    print "Currently, it points to:"
    print BOOST_BASE
    print 70*"-"
    sys.exit(1)

# Check if the supportlib has already been compiled...
if sys.platform=="win32":
    libcore = "core.lib"
else:
    libcore = "libcore.a"
libcore = os.path.join("supportlib", "lib", libcore)
if not os.path.exists(libcore) and not INSTALL_CGKIT_LIGHT:
    print 70*"-"
    print "The library %s does not exist."%libcore
    print "You have to compile the C++ support library first. Please go to the"
    print "subdirectory 'supportlib' and invoke scons:\n"
    print " > cd supportlib"
    print " > scons\n"
    print "After that, you can proceed with calling the setup script. Please"
    print "see the file readme.txt or the section about compiling the package"
    print "in the manual if you encounter any further problems."
    print "Note: You can read the manual online at http://cgkit.sf.net"
    print 70*"-"
    sys.exit(1)

##### Windows specific stuff #####
if sys.platform=="win32":

    MACROS += [("WIN32", None)]

    # OpenGL
    LIBS += ["opengl32", "glu32"]

    # Additional compiler options
    CC_ARGS  += ["/GR", "/Zm500"]
    # Additional linker options
    LINK_ARGS += ["/NODEFAULTLIB:LIBCMT"]

    # Add the boost dll as additional file
    # (it is supposed to go into the root package directory)
    if BOOST_BIN!=None:
        boost_dll = os.path.join(BOOST_BIN, BOOST_DLL)
        packageDLL(boost_dll)

    # Add the stlport dll as additional file
    # (it is supposed to go into the root package directory)
    if USING_STLPORT:
        if os.path.isabs(STLPORT_DLL):
            stlport_dll = STLPORT_DLL
        else:
            stlport_dll = os.path.join(STLPORT_BASE, "lib", STLPORT_DLL)
        packageDLL(stlport_dll)

    # Lib3ds
    if LIB3DS_AVAILABLE:
        INC_DIRS += [LIB3DS_PATH]
        LIB_DIRS += [os.path.join(LIB3DS_PATH, "msvc", "Build", "Release")]
#        LIBS += ["lib3ds-120s"]
#        MACROS += [("LIB3DS_AVAILABLE", None)]

    # CyberX3D
    if CYBERX3D_AVAILABLE:
        INC_DIRS += [os.path.join(CYBERX3D_PATH, "include")]
        LIB_DIRS += [os.path.join(CYBERX3D_PATH, "lib")]
        LIB_DIRS += [os.path.join(XERCES_PATH, "lib")]
	if CYBERX3D_LIB==None:
	    CYBERX3D_LIB = "CyberX3D"
	if CYBERX3D_XERCES_LIB==None:
	    CYBERX3D_XERCES_LIB = "xerces-c_2"
        LIBS += [CYBERX3D_LIB, CYBERX3D_XERCES_LIB, "gdi32"]
        if os.path.isabs(CYBERX3D_XERCES_DLL):
            packageDLL(CYBERX3D_XERCES_DLL)
        else:
            packageDLL(os.path.join(XERCES_PATH, "bin", CYBERX3D_XERCES_DLL))

##### Mac OS X (Darwin) #####
elif sys.platform=="darwin":
    # OpenGL 
    INC_DIRS += ["/System/Library/Frameworks/OpenGL.framework/Versions/A/Headers/"]
    LINK_ARGS += ["-framework AGL -framework OpenGL"]

    # CyberX3D
    if CYBERX3D_AVAILABLE:
        if CYBERX3D_LIB==None:
            CYBERX3D_LIB = "cx3d"
        if CYBERX3D_XERCES_LIB==None:
            CYBERX3D_XERCES_LIB = "xerces-c"
        LIBS += [CYBERX3D_LIB, CYBERX3D_XERCES_LIB]
   
##### Linux (and other) #####
else:
    # OpenGL 
    LIBS += ["GL", "GLU"]

    # CyberX3D
    if CYBERX3D_AVAILABLE:
	if CYBERX3D_LIB==None:
	    CYBERX3D_LIB = "cx3d"
	if CYBERX3D_XERCES_LIB==None:
	    CYBERX3D_XERCES_LIB = "xerces-c"
        LIBS += [CYBERX3D_LIB, CYBERX3D_XERCES_LIB]

    
######################################################################

core_src_files = ["wrappers/py_wrapper.cpp",
                  "wrappers/py_noise.cpp",
                  "wrappers/py_vec3.cpp",
                  "wrappers/py_vec4.cpp",
                  "wrappers/py_mat3.cpp",
                  "wrappers/py_mat4.cpp",
                  "wrappers/py_quat.cpp",
                  "wrappers/py_slots1.cpp",
                  "wrappers/py_slots2.cpp",
                  "wrappers/py_slots3.cpp",
                  "wrappers/py_slots4.cpp",
                  "wrappers/py_arrayslots1.cpp",
                  "wrappers/py_arrayslots2.cpp",
                  "wrappers/py_arrayslot_mat3.cpp",
                  "wrappers/py_arrayslot_mat4.cpp",
                  "wrappers/py_arrayslots4.cpp",
                  "wrappers/py_component.cpp",
                  "wrappers/py_boundingbox.cpp",
                  "wrappers/py_worldobject.cpp",
                  "wrappers/py_worldobject2.cpp",
                  "wrappers/py_worldobject3.cpp",
                  "wrappers/py_material.cpp",
                  "wrappers/py_glmaterial.cpp",
                  "wrappers/py_geoms1.cpp",
                  "wrappers/py_spheregeom.cpp",
                  "wrappers/py_ccylindergeom.cpp",
                  "wrappers/py_torusgeom.cpp",
                  "wrappers/py_boxgeom.cpp",
                  "wrappers/py_planegeom.cpp",
                  "wrappers/py_trimeshgeom.cpp",
                  "wrappers/py_polyhedrongeom.cpp",
                  "wrappers/py_drawgeom.cpp",
                  "wrappers/py_lightsource.cpp",
                  "wrappers/py_glpointlight.cpp",
                  "wrappers/py_glspotlight.cpp",
                  "wrappers/py_gldistantlight.cpp",
                  "wrappers/py_glrenderer.cpp",
                  "wrappers/rply/rply/rply.c",
                  "wrappers/rply/py_rply_read.cpp",
                  "wrappers/rply/py_rply_write.cpp"]

# Shall the lib3ds wrappers be included?
if LIB3DS_AVAILABLE:
    core_src_files += ["wrappers/lib3ds/common_funcs.cpp",
                       "wrappers/lib3ds/nodedata.cpp",
                       "wrappers/lib3ds/light.cpp",
                       "wrappers/lib3ds/mesh.cpp",
                       "wrappers/lib3ds/material.cpp",
                       "wrappers/lib3ds/lib3ds.cpp"]

# Shall the CyberX3D wrappers be included?
if CYBERX3D_AVAILABLE:
    core_src_files += ["wrappers/cyberx3d/cyberx3d.cpp"]

# Shall the ogre wrappers be included?
if OGRE_AVAILABLE:
    core_src_files += ["wrappers/ogre/OgreCore.cpp",
                       "wrappers/ogre/OgreWrap.cpp"]


# Extension modules:

# Core library
ext_modules = [Extension("_core", core_src_files
                         ,extra_link_args=LINK_ARGS
                         ,libraries=LIBS
                         ,include_dirs=INC_DIRS
                         ,library_dirs=LIB_DIRS
                         ,extra_compile_args=CC_ARGS
                         ,define_macros=MACROS)]

# 3DXWare SpaceMouse/SpaceBall
if THREEDXWARE_AVAILABLE:
    ext_modules += [Extension("_spacedevice", ["wrappers/spacedevice/py_spacedevice.cpp"]
                             ,libraries=LIBS
                             ,include_dirs=INC_DIRS
                             ,library_dirs=LIB_DIRS
                             ,extra_compile_args=CC_ARGS
                             ,extra_link_args=LINK_ARGS
                             ,define_macros=MACROS)]

# Wintab
if WINTAB_AVAILABLE:
    ext_modules += [Extension("_wintab", ["wrappers/wintab/py_wintab.cpp"]
                             ,libraries=LIBS
                             ,include_dirs=INC_DIRS
                             ,library_dirs=LIB_DIRS
                             ,extra_compile_args=CC_ARGS
                             ,extra_link_args=LINK_ARGS
                             ,define_macros=MACROS)]

# Data glove
if GLOVESDK_AVAILABLE:
    if GLOVESDK_BASE!=None:
        INC_DIRS += [os.path.join(GLOVESDK_BASE, "inc")]
        LIB_DIRS += [os.path.join(GLOVESDK_BASE, "lib")]

    if sys.platform=="win32":
        MACROS += [("GLOVESDK_2_0", None)]
        if os.path.isabs(GLOVESDK_DLL):
            packageDLL(GLOVESDK_DLL)
        else:
            packageDLL(os.path.join(GLOVESDK_BASE, "dll",  GLOVESDK_DLL))
        
    ext_modules += [Extension("_glove", ["wrappers/glove/py_glove.cpp"]
                             ,libraries=LIBS
                             ,include_dirs=INC_DIRS
                             ,library_dirs=LIB_DIRS
                             ,extra_compile_args=CC_ARGS
                             ,extra_link_args=LINK_ARGS
                             ,define_macros=MACROS)]


# Infos...
updateInfoModule(INSTALL_CGKIT_LIGHT)
print 70*"="
print "3DS importer:      %s"%(enabledStr(LIB3DS_AVAILABLE))
print "VRML/X3D importer: %s"%(enabledStr(CYBERX3D_AVAILABLE))
print "OGRE:              %s"%(enabledStr(OGRE_AVAILABLE))
print "3DXWare:           %s"%(enabledStr(THREEDXWARE_AVAILABLE))
print "Wintab:            %s"%(enabledStr(WINTAB_AVAILABLE))
print "Glove module:      %s"%(enabledStr(GLOVESDK_AVAILABLE))
print 70*"="

# Check if config.cpp is newer than py_wrapper.cpp
# If it is, "touch" py_wrapper.cpp so that it will be compiled again
# (because optional libs might have been included or excluded)
if os.path.exists("config.cfg") and not INSTALL_CGKIT_LIGHT:
    if isNewer("config.cfg", "wrappers/py_wrapper.cpp"):
        print "Forcing compilation of py_wrapper.cpp"
        os.utime("wrappers/py_wrapper.cpp", None)


# Test (to enable light version)
if INSTALL_CGKIT_LIGHT:
    ext_modules = []

setup(name = PACKAGE_NAME,
      version = "2.0.0alpha5",
      description = "Python Computer Graphics Kit",
      author = "Matthias Baas",
      author_email = "baas@ira.uka.de",
      url = "http://cgkit.sourceforge.net",
      license = "LGPL",

      # Process all pure Python modules in the cgkit directory
      packages = [PACKAGE_NAME,
                  PACKAGE_NAME+"._OpenGL",
                  PACKAGE_NAME+".all",
                  PACKAGE_NAME+".light",
                  PACKAGE_NAME+".light.cgtypes",
                  PACKAGE_NAME+".GUI",
                  PACKAGE_NAME+".Interfaces",
                  PACKAGE_NAME+".wintab",
                  PACKAGE_NAME+".spacedevice",
                  PACKAGE_NAME+".glove"],

      # Stick all extension modules into the main package
      ext_package = PACKAGE_NAME,
      ext_modules = ext_modules,
      scripts = scripts,
      data_files = data_files
      )

