/*======================================================================
 cgkit - Python Computer Graphics Kit
 Copyright (C) 2004 Matthias Baas (baas@ira.uka.de)

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 cgkit homepage: http://cgkit.sourceforge.net
======================================================================*/

#include "glmaterial.h"

#ifdef WIN32
#include <windows.h>
#endif

#if defined(__APPLE__) || defined(MACOSX)
#include "OpenGL/gl.h"
#include <OpenGL/glu.h>
#else
#include "GL/gl.h"
#include <GL/glu.h>
#endif

namespace support3d {

//////////////////////////////////////////////////////////////
// Texture
//////////////////////////////////////////////////////////////

/**
  Constructor.
 */
GLTexture::GLTexture()
: environment_map(false),
  mipmap(true),
  mode(GL_DECAL),
  mag_filter(GL_LINEAR),
  min_filter(GL_LINEAR),
  wrap_s(GL_REPEAT),
  wrap_t(GL_REPEAT),
  internalformat(GL_RGB),
  texenvcolor(1),
  transform(1),
  imagename(),
  texname(0),
  flags(0)
{
}

/**
  Destructor.
 */
GLTexture::~GLTexture() 
{
}

/**
  Return the name of the texture image.

  \return Image name
 */
string GLTexture::getImageName()
{
  return imagename;
}

/**
  Set a new texture image name.

  The texture will be reloaded in the next call to applyGL().

  \param name New texture image name
 */
void GLTexture::setImageName(string name)
{
  imagename = name;
  flags |= IMAGE_NAME;
}

/**
  Do the OpenGL calls to activate the texture.
 */
void GLTexture::applyGL()
{
  if (texname==0)
  {
    allocGL();
  }

  glBindTexture(GL_TEXTURE_2D, texname);

  // Has the image name changed? Then reload
  if (flags & IMAGE_NAME)
  {
    flags &= ~IMAGE_NAME;
    loadTexData();
  }

  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, mode);
  GLfloat col[4] = {GLfloat(texenvcolor.x), GLfloat(texenvcolor.y), 
                    GLfloat(texenvcolor.z), GLfloat(texenvcolor.w)};
  glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, col);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mag_filter);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min_filter);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap_s);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap_t);

  glMatrixMode(GL_TEXTURE);
  double M[16];
  glLoadIdentity();
  transform.toList(M);
  glMultMatrixd(M);
  glMatrixMode(GL_MODELVIEW);

  if (environment_map)
  {
    // Enable environment mapping
    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
  }
  else
  {
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
  }
}

/**
  Allocate the OpenGL texture object.
 */
void GLTexture::allocGL()
{
  GLuint tn[1];

  // Alloc texture name and create texture object...
  glGenTextures(1, tn);
  texname = tn[0];
  glBindTexture(GL_TEXTURE_2D, texname);

  std::cout<<"Texname: "<<texname<<std::endl;
}

void GLTexture::releaseGL()
{
}

/**
  Pass the texture data to OpenGL.

  This method has to be called by the texData() method.

  \param w Width of the texture image (must be a power of 2)
  \param h Height of the texture image (must be a power of 2)
  \param format Format of the data (see format parameter of the glTexImage2D() function)
  \param type Type of the data (see type parameter of the glTexImage2D() function)
  \param data Image data
 */
void GLTexture::texData(int w, int h, int format, int type, char* data)
{
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  if (mipmap)
    gluBuild2DMipmaps(GL_TEXTURE_2D, internalformat, w, h, format, type, data);
  else
    glTexImage2D(GL_TEXTURE_2D, 0, internalformat, w, h, 0, format, type, data);
}


//////////////////////////////////////////////////////////////
// Material
//////////////////////////////////////////////////////////////

/**
  Constructor.
 */
GLMaterial::GLMaterial()
 : Material(),
   ambient(vec4d(0.2,0.2,0.2,1), 0),
   diffuse(vec4d(0.7,0.7,0.7,1), 0),
   specular(vec4d(0,0,0,1), 0),
   shininess(0.0, 0),
   emission(vec4d(0,0,0,1), 0),
   blend_sfactor(-1),
   blend_dfactor(-1)
{ 
  addSlot("ambient", ambient);
  addSlot("diffuse", diffuse);
  addSlot("specular", specular);
  addSlot("shininess", shininess);
  addSlot("emission", emission);
}

/**
  Constructor.
 */
GLMaterial::GLMaterial(string aname, double adensity)
 : Material(aname, adensity),
   ambient(vec4d(0.2,0.2,0.2,1), 0),
   diffuse(vec4d(0.7,0.7,0.7,1), 0),
   specular(vec4d(0,0,0,1), 0),
   shininess(0.0, 0),
   emission(vec4d(0,0,0,1), 0),
   blend_sfactor(-1),
   blend_dfactor(-1),
   texture()
{ 
  addSlot("ambient", ambient);
  addSlot("diffuse", diffuse);
  addSlot("specular", specular);
  addSlot("shininess", shininess);
  addSlot("emission", emission);
}

/** Destructor. */
GLMaterial::~GLMaterial()
{
}

/**
  Apply the material using OpenGL commands.

  The material settings are only applied to the front faces.

  \pre The appropriate OpenGL context has been made current
 */
void GLMaterial::applyGL()
{
  GLfloat c[4] = {0,0,0,1};
  GLfloat f;
  
  const vec4d& ac = ambient.getValue();
  c[0] = GLfloat(ac.x);
  c[1] = GLfloat(ac.y);
  c[2] = GLfloat(ac.z);
  c[3] = GLfloat(ac.w);
  glMaterialfv(GL_FRONT, GL_AMBIENT, c);

  const vec4d& dc = diffuse.getValue();
  c[0] = GLfloat(dc.x);
  c[1] = GLfloat(dc.y);
  c[2] = GLfloat(dc.z);
  c[3] = GLfloat(dc.w);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, c);

  const vec4d& sc = specular.getValue();
  c[0] = GLfloat(sc.x);
  c[1] = GLfloat(sc.y);
  c[2] = GLfloat(sc.z);
  c[3] = GLfloat(sc.w);
  glMaterialfv(GL_FRONT, GL_SPECULAR, c);

  const vec4d& ec = emission.getValue();
  c[0] = GLfloat(ec.x);
  c[1] = GLfloat(ec.y);
  c[2] = GLfloat(ec.z);
  c[3] = GLfloat(ec.w);
  glMaterialfv(GL_FRONT, GL_EMISSION, c);

  f = GLfloat(shininess.getValue());
  glMaterialf(GL_FRONT, GL_SHININESS, f);

  if (texture.get()!=0)
  {
    texture->applyGL();
    glEnable(GL_TEXTURE_2D);
  }
  else
  {
    glDisable(GL_TEXTURE_2D);
  }

  // Blending?
  if (usesBlending())
  {
    glEnable(GL_BLEND);
    glBlendFunc(blend_sfactor, blend_dfactor);
  }
  else
  {
    glDisable(GL_BLEND);
  }
}

/**
  Return True if the material activates blending.
 */
bool GLMaterial::usesBlending()
{
  return (blend_sfactor!=-1) && (blend_dfactor!=-1);
}

}  // end of namespace
